# ## What are Machines?
# {Machine} and {MachineInput} together implement a kind of observer pattern.
# {Machine} processes a collection of input cards to generate an output card
# (a {Set::Type::File} card by default). If one of the input cards is changed
# the output card will be updated.
#
# The classic example: A style card observes a collection of css and sccs card
# to generate a file card with a css file that contains the assembled
# compressed css.
#
# ## Using Machines
# Include the Machine module in the card set that is supposed to produce the
# output card. If the output card should be automatically updated when a input
# card is changed the input card has to be in a set that includes the
# MachineInput module.
#
# The default machine:
#
#  -  uses its item cards as input cards or the card itself if there are no
# item cards;
#  -  can be changed by passing a block to collect_input_cards
#  -  takes the raw view of the input cards to generate the output;
#  -  can be changed by passing a block to machine_input (in the input card
# set)
#  -  stores the output as a .txt file in the '+machine output' card;
#  -  can be changed by passing a filetype and/or a block to
#     store_machine_output
#
#
# ## How does it work?
# Machine cards have a '+machine input' and a '+machine output' card. The
# '+machine input' card is a pointer to all input cards. Including the
# MachineInput module creates an 'on: save' event that runs the machines of
# all cards that are linked to that card via the +machine input pointer.
module MachineClassMethods
  attr_accessor :output_config

  def collect_input_cards &block
    define_method :engine_input, &block
  end

  def prepare_machine_input &block
    define_method :before_engine, &block
  end

  def machine_engine &block
    define_method :engine, &block
  end

  def store_machine_output args={}, &block
    output_config.merge!(args)
    return unless block_given?

    define_method :after_engine, &block
  end
end

card_accessor :machine_output, type: FileID
card_accessor :machine_input, type: PointerID

def before_engine; end

def engine_input
  ei = EngineInput.new self
  ei.process
  ei.new_input
end

# TODO: replace with call of extended_item_cards
# traverse through all levels of pointers and
# collect all item cards as input
class EngineInput
  attr_accessor :new_input

  def initialize machine_card
    @machine_card = machine_card
    @items = [machine_card]
    @new_input = []
    @extended = {}
    @loop_limit = 5
  end

  def process
    each_valid_item do
      input_item = simple_item? ? @item : pointer_item
      new_input << input_item if input_item
    end
  end

  def simple_item?
    @item.item_cards == [@item] || @item.try(:self_machine_input?) # no pointer card
  end

  def pointer_item
    @items.insert 0, @item.item_cards.reject(&:unknown?)
    @items.flatten!
    record_item
    @item if @item != @machine_card && @item.known?
  end

  def record_item
    @extended[@item] = @extended[@item].to_i + 1
  end

  def each_valid_item
    until @items.empty?
      @item = @items.shift
      yield unless invalid_item?
    end
  end

  def invalid_item?
    @item.trash || @extended[@item.id].to_i > @loop_limit
  end
end

def engine input
  input
end

def after_engine output
  filetype = output_config[:filetype]
  file = Tempfile.new [id.to_s, ".#{filetype}"]
  file.write output
  file.rewind
  store_machine_output file
  file.close
  file.unlink
end

format do
  view :machine_output_url do
    machine_output_url
  end
end

class << self
  def included host_class
    host_class.extend(MachineClassMethods)
    host_class.mattr_accessor :output_config
    host_class.output_config = { filetype: "txt" }

    define_machine_events host_class
  end

  def define_machine_events host_class
    event_suffix = host_class.name.tr ":", "_"
    event_name = "reset_machine_output_#{event_suffix}".to_sym
    host_class.event event_name, after: :expire_related, changed: :content, on: :save do
      reset_machine_output
    end
  end
end

include_set Abstract::Lock

def run_machine joint="\n"
  before_engine
  output =
    input_item_cards.map do |input_card|
      run_engine input_card
    end.select(&:present?).join(joint)
  after_engine output
end

def direct_machine_input? input_card
  !input_card.collection? ||
    input_card.respond_to?(:machine_input)
end

def run_engine input_card
  return unless direct_machine_input? input_card
  # if (cached = fetch_cache_card(input_card)) && cached.content?
  #  return cached.content
  # end

  engine(input_from_card(input_card)).tap do |output|
    cache_output_part input_card, output
  end
end

def input_from_card input_card
  if input_card.respond_to? :machine_input
    input_card.machine_input
  else
    input_card.format._render_raw
  end
end

# FIXME: this is wack. set the storage right the first time through. obviate
def make_machine_output_coded mod=:machines
  update_machine_output
  Card::Auth.as_bot do
    ENV["STORE_CODED_FILES"] = "true"
    machine_output_card.update! storage_type: :coded, mod: mod,
                                codename: machine_output_codename
    ENV["STORE_CODED_FILES"] = nil
  end
end

def machine_output_codename
  machine_output_card.name.parts.map do |part|
    Card[part].codename&.to_s || Card[part].name.safe_key
  end.join "_"
end

def input_item_cards
  machine_input_card.item_cards
end

def machine_output_url
  ensure_machine_output
  machine_output_card.file.url # (:default, timestamp: false)
  # to get rid of additional number in url
end

def machine_output_path
  ensure_machine_output
  machine_output_card.file.path
end

private

# TODO: make more methods private!

def store_machine_output file
  Card::Auth.as_bot do
    moc = machine_output_card
    prepare_coded_output moc
    moc.file = file
    moc.save!
  end
end

def prepare_coded_output moc
  return unless (dir = ENV["SEED_MACHINE_OUTPUT_TO"])
  ENV["STORE_CODED_FILES"] = "true"
  moc.storage_type = moc.new_storage_type = :coded
  moc.mod = dir
  moc.codename = machine_output_codename
end
